package cks;

import cks.parser.*;
import cks.generator.*;
import cks.model.*;
import cks.model.type.Type;
import cks.model.type.TNamed;
import cks.model.value.Value;

import java.util.Map;
import java.util.HashMap;
import java.io.PrintWriter;
import java.io.FileReader;
import java.io.IOException;
import java.io.File;

public final class Main
{
	enum Stage { Parse, Resolve, Generate, }

	private static class Exit extends Error
	{
		private final int code;
		public Exit(int code)
		{
			this.code = code;
		}
	}

	private static Exit exit(int code)
	{
		return new Exit(code);
	}

  public static void main(String[] args)
		throws IOException
  {
		try {
			PrintWriter out = new PrintWriter(System.out);
			PrintWriter err = new PrintWriter(System.err);

			try {
				main_(args, out, err);
			} finally {
				out.flush();
				err.flush();
			}

		} catch (Exit e) {
			System.exit(e.code);
		}
	}

  public static void main_(String[] args, PrintWriter out, PrintWriter err)
		throws IOException
  {
		if (args.length == 0) {
			printUsage(err);
			throw exit(1);
		}

		String mode = args[0];

		if (mode.equals("type")) {
			parseType(out, err, args);
		}
		else if (mode.equals("value")) {
			parseValue(out, err, args);
		}
		else {
			printUsage(err);
			throw exit(1);
		}
	}

	private static void parseType(PrintWriter out, PrintWriter err, String[] args)
		throws IOException
	{
		if (args.length != 3) {
			printUsage(err);
			throw exit(1);
		}

		String stageString = args[1];
		Stage stage;

		if (stageString.equals("parse")) {
			stage = Stage.Parse;
		}
		else if (stageString.equals("resolve")) {
			stage = Stage.Resolve;
		}
		else if (stageString.equals("gen")) {
			stage = Stage.Generate;
		} else {
			System.err.println("Bad value for 'stage': \"" + stageString + "\" (valid values: parse, resolve, gen)");
			throw exit(1);
		}

		File inputFile = new File(args[2]);

		// Take everything before the dot
		String typeName = inputFile.getName();
		{
			int dotPos = typeName.indexOf('.');
			if (dotPos >= 0) typeName = typeName.substring(0, dotPos);
		}

		FileReader reader = new FileReader(inputFile);
		try {
			parseType(typeName, new StreamTokenizer(reader, inputFile.getPath()), out, stage);
			out.flush();
		} catch (IOException ex) {
			System.err.println("Error reading from \"" + inputFile.getPath() + "\": " + ex);
			throw exit(1);
		} catch (LocatableException ex) {
			ex.printStackTrace(System.err);
		} finally {
			close(err, inputFile, reader);
		}

  }

	public static void printUsage(PrintWriter out)
	{
		out.println("Usage:");
		out.println("  COMMAND type ( parse | resolve | gen ) <type-file>");
		out.println("  COMMAND value <type-file> <value-file>");
	}

  public static void parseType(String name, TokenSource ts, PrintWriter out, Stage stage)
    throws LocatableException, IOException
  {
    Type type = SimpleTypeParser.parse(ts);

		if (stage == Stage.Parse) {
			type.dump(out);
			out.flush();
			return;
		}

		type.resolve(Type.STANDARD_CONTEXT);

		type.dump(out);
		out.flush();

		if (stage == Stage.Resolve) return;

		JavaGenerator.generate(out, type, "test", name);
		out.flush();

		if (stage == Stage.Generate) return;

		assert false : stage;
	}

	private static void parseValue(PrintWriter out, PrintWriter err, String[] args)
		throws IOException
	{

		if (args.length != 3) {
			printUsage(err);
			throw exit(1);
		}

		File typeFile = new File(args[1]);
		File valueFile = new File(args[2]);

		FileReader tfr = new FileReader(typeFile);
		FileReader vfr = new FileReader(valueFile);
		try {
			TokenSource tts = new StreamTokenizer(tfr, typeFile.getPath());
			TokenSource vts = new StreamTokenizer(vfr, valueFile.getPath());
			parseValue(tts, vts, out);
			out.flush();
		} catch (IOException ex) {
			err.println("Error reading from either \"" + typeFile.getPath() + "\" or \"" + valueFile.getPath() + "\": " + ex);
			throw exit(1);
		} catch (LocatableException ex) {
			ex.printStackTrace(System.err);
		} finally {
			close(err, typeFile, tfr);
			close(err, valueFile, vfr);
		}

  }

	private static void close(PrintWriter out, File file, FileReader reader)
	{
		try {
			reader.close();
		} catch (IOException ex) {
			out.println("Error closing input file \"" + file.getPath() + "\": " + ex);
		}
	}

	private static void parseValue(TokenSource tts, TokenSource vts, PrintWriter out)
    throws LocatableException, IOException
	{
    Type type = SimpleTypeParser.parse(tts);

		type.resolve(Type.STANDARD_CONTEXT);

		Value value = SimpleValueParser.parse(vts, type);
		value.dump(out);
	}
}
