package cks.generator;

import java.io.Writer;
import java.io.PrintWriter;
import java.io.PrintStream;

public final class IndentPrinter extends PrintWriter
{
	private int currentSpaces;
	private int indentSize;

	// Tracks whether the last call was a 'println' and we need
	// to print indentation before outputting the next character.
	private boolean needToPrintIndent = true;

	private static final int DEFAULT_INDENT_SIZE = 2;

	/**
	 * @param indentSize
	 *    The number of spaces to use for each indentation increase.
	 *
	 * @param startSpaces
	 *    The number of spaces of indentation to start out with.
	 */
	public IndentPrinter(Writer out, int indentSize, int startSpaces)
	{
		super(out);
		this.indentSize = indentSize;
		this.currentSpaces = startSpaces;
	}

	public IndentPrinter(Writer out, int indentSize)
	{
		this(out, indentSize, 0);
	}

	public IndentPrinter(Writer out)
	{
		this(out, DEFAULT_INDENT_SIZE);
	}

	/**
	 * @param indentSize
	 *    The number of spaces to use for each indentation increase.
	 *
	 * @param startSpaces
	 *    The number of spaces of indentation to start out with.
	 */
	public IndentPrinter(PrintStream out, int indentSize, int startSpaces)
	{
		super(out);
		this.indentSize = indentSize;
		this.currentSpaces = startSpaces;
	}

	public IndentPrinter(PrintStream out, int indentSize)
	{
		this(out, indentSize, 0);
	}

	public IndentPrinter(PrintStream out)
	{
		this(out, DEFAULT_INDENT_SIZE);
	}

	// -----------------------------------------
	// indentation control

	/**
	 * Increase the indentation level by 'amount' indentation.s
	 */
	public void indent(int amount)
	{
		currentSpaces += amount * indentSize;
	}

	public void indent()
	{
		currentSpaces += indentSize;
	}

	public void dedent(int amount)
	{
		currentSpaces -= amount * indentSize;
	}

	public void dedent()
	{
		currentSpaces -= indentSize;
	}

	// -----------------------------------------
	// print functions

	public void print(String v)
	{
		prefix();
		super.print(v);
	}
	public void print(char[] v)
	{
		prefix();
		super.print(v);
	}
	public void print(Object v)
	{
		prefix();
		super.print(v);
	}
	public void print(char v)
	{
		prefix();
		super.print(v);
	}
	public void print(int v)
	{
		prefix();
		super.print(v);
	}
	public void print(long v)
	{
		prefix();
		super.print(v);
	}
	public void print(double v)
	{
		prefix();
		super.print(v);
	}
	public void print(float v)
	{
		prefix();
		super.print(v);
	}
	public void print(boolean v)
	{
		prefix();
		super.print(v);
	}

	public IndentPrinter append(CharSequence v)
	{
		prefix();
		super.append(v);
		return this;
	}
	public IndentPrinter append(CharSequence v, int start, int end)
	{
		prefix();
		super.append(v, start, end);
		return this;
	}
	public IndentPrinter append(char c)
	{
		prefix();
		super.append(c);
		return this;
	}

	// -----------------------------------------
	// print-ln functions

	public void println(String v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(char[] v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(Object v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(char v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(int v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(long v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(double v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(float v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println(boolean v)
	{
		prefix();
		super.println(v);
		needToPrintIndent = true;
	}
	public void println()
	{
		prefix(); // <--- if you remove this, blank lines will not be indented
		super.println();
		needToPrintIndent = true;
	}

	// -----------------------------------------
	// formatted print crap
	
	public IndentPrinter format(String format, Object... args)
	{
		prefix();
		super.format(format, args);
		return this;
	}

	public IndentPrinter format(java.util.Locale l, String format, Object... args)
	{
		prefix();
		super.format(l, format, args);
		return this;
	}

	public IndentPrinter printf(String format, Object... args)
	{
		return format(format, args);
	}

	public IndentPrinter printf(java.util.Locale l, String format, Object... args)
	{
		return format(l, format, args);
	}

	// -----------------------------------------
	// raw write functions
	
	public void write(char[] buf)
	{
		prefix();
		super.write(buf);
	}

	public void write(char[] buf, int off, int len)
	{
		prefix();
		super.write(buf, off, len);
	}

	public void write(char c)
	{
		prefix();
		super.write(c);
	}

	public void write(String s)
	{
		prefix();
		super.write(s);
	}

	public void write(String s, int off, int len)
	{
		prefix();
		super.write(s, off, len);
	}

	// -----------------------------------------
	// indentation functions

	private void prefix()
	{
		if (needToPrintIndent) {
			printIndent();
			needToPrintIndent = false;
		}
	}

	private static final char[] SPACES = {
		' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
		' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',
	};

	private void printIndent()
	{
		int remainingSpaces = this.currentSpaces;
		while (remainingSpaces > SPACES.length) {
			super.write(SPACES);
			remainingSpaces -= SPACES.length;
		}
		super.write(SPACES, 0, remainingSpaces);
	}

}
