package cks.generator;

import cks.model.*;
import cks.model.type.*;
import cks.parser.*;

import java.io.PrintWriter;
import java.io.IOException;
import java.io.Writer;
import java.io.FileReader;
import java.io.File;
import java.util.Map;
import java.util.TreeMap;
import java.util.Collection;

public class JavaGenerator
{
  public static class Exception extends ModelException
  {
    public Exception(Node node, String message)
    {
      super(node, message);
    }
  }

  public static void generate(Writer outRaw, Type type, String packageName, String name)
		throws Exception
	{
    assert(name != null);

		JavaGenerator gen = new JavaGenerator(new IndentPrinter(outRaw), packageName, name);

		gen.printHeader();
		gen.gen(name, type, null);
  }

	private JavaGenerator(IndentPrinter out, String packageName, String sourceFile)
	{
		this.out = out;
		this.packageName = packageName;
		this.sourceFile = sourceFile;
	}

	private final IndentPrinter out;
	private final String packageName;
	private final String sourceFile;

	private String gen(String name, Type type)
		throws Exception
	{
		return gen(name, type, null, false);
	}

	private String gen(String name, Type type, Collection<TParam> params)
		throws Exception
	{
		return gen(name, type, params, false);
	}

	private String gen(String name, Type type, Collection<TParam> params, boolean top)
		throws Exception
	{
		if (type instanceof TCompound) {
			return gen(name, (TCompound) type, params, null, top);
		}
		else if (type instanceof TIndirect) {
			return getRef((TIndirect) type);
		}
		else {
			throw badType(type);
		}
	}

	private String gen(String name, TCompound type, Collection<TParam> params, String groupRef)
		throws Exception
	{
		return gen(name, type, params, groupRef, false);
	}

	private String gen(String name, TCompound type, Collection<TParam> params, String groupRef, boolean top)
		throws Exception
	{
		if (type instanceof TRecord) {
			return gen(name, (TRecord) type, params, groupRef, top);
		}
		else if (type instanceof TCompound) {
			return gen(name, (TVariant) type, params, groupRef, top);
		}
		else {
			throw badType(type);
		}
	}

	private String gen(String name, TRecord type, Collection<TParam> params, String groupRef, boolean top)
		throws Exception
	{
		// Prerequisites
		
		String parentRef = null;
		if (type.parentRef != null) {
			parentRef = getRef(type.parentRef);
		}

		// Content

		if (top) {
			out.print("public static class ");
		} else {
			out.print("public class ");
		}
		out.print(name + genParamDecl(params));

		if (parentRef != null) {
			out.print(" extends " + parentRef);
		}

		if (groupRef != null) {
			out.print(" implements " + groupRef);
		}

		out.println();

		braceOpen();

		for (TMember field : type.members.values()) {
			String fieldTypeRef = gen(field.name, field.type);
			out.println("private " + fieldTypeRef + " _" + field.name + ";");
		}

		gen(type.defs);

		braceClose();

		return name;
	}

	private String gen(String name, TVariant type, Collection<TParam> params, String groupRef, boolean top)
		throws Exception
	{
		// Prerequisites
		
		String naturalParentRef = null;
		if (type.parentRef != null) {
			naturalParentRef = getRef(type.parentRef);
		}

		// Content

		if (top) {
			out.print("public static abstract interface ");
		} else {
			out.print("public abstract interface ");
		}

		String paramDecl = genParamDecl(params);
		String myRef = name + paramDecl;
		out.print(myRef);

		if (groupRef != null) {
			out.print(" extends " + groupRef);
		}

		out.println();

		braceOpen();

		for (TMember field : type.members.values()) {
			if (field.type instanceof TCompound) {
				// If the field type is a compound definition, we can impose our "implements"
				// constraint directly on it.  This means we don't need to create a container
				// class.
				String fieldTypeRef = gen(field.name, (TCompound) field.type, params, myRef);
			}
			else if (field.type instanceof TIndirect) {
				// If it's not a compound definition, create a container class with a single
				// "value" field.
				TIndirect ind = (TIndirect) field.type;
				out.println("private static class " + field.name + paramDecl + " implements " + myRef);
				braceOpen();
				out.println("public final " + getRef(ind) + " value;");
				braceClose();
			}
			else {
				throw badType(field.type);
			}
		}

		gen(type.defs);

		braceClose();

		return name;
	}

	private void gen(Map<String,TDef> defs)
		throws Exception
	{
		for (TDef def : defs.values()) {
			gen(def);
		}
	}

	private String gen(TDef def)
		throws Exception
	{
		if (def instanceof TDef.Simple) {
			TDef.Simple simple = (TDef.Simple) def;
			return gen(def.name, def.value, null);
		}
		else if (def instanceof TDef.Function) {
			TDef.Function func = (TDef.Function) def;
			return gen(func.name, func.value, func.params.values());
		}
		else {
			throw badType(def);
		}
	}

	private void braceOpen()
	{
		out.println("{");
		out.indent();
	}

	private void braceClose()
	{
		out.dedent();
		out.println("}");
	}

	private int freshVar = 0;

	private String getRef(TIndirect type)
		throws Exception
	{
		if (type instanceof TRef) {
			return getRef((TRef) type);
		}
		else if (type instanceof TInvoke) {
			return getRef((TInvoke) type);
		}
		else if (type instanceof TOpaqueRef) {
			return getRef((TOpaqueRef) type);
		}
		else {
			throw badType(type);
		}
	}

	private String getRef(TRef type)
		throws Exception
	{
		return getRef(type.target);
	}

	private int unique = 0;

	private String getRef(TInvoke type)
		throws Exception
	{
		if (!type.getKind().equals(TKind.BASE)) throw new Exception(type, "can't handle partial application");

		StringBuilder buf = new StringBuilder();

		// base type
		buf.append(getRef(type.target));

		// type arguments
		assert !type.args.isEmpty();
		buf.append('<');
		for (Type arg : type.args) {
			String ref = gen("Temp_" + (unique++), arg);
			buf.append(ref);
			buf.append(",");
		}
		buf.setLength(buf.length() - 1);
		buf.append('>');
		return buf.toString();
	}

	private String getRef(TOpaqueRef type)
		throws Exception
	{
		String ref = getRef(type.target);
		if (ref == null) {
			throw new Exception(type, "can't handle opaque type \"" + type.target.name + "\"");
		}
		return ref;
	}

	private String getRef(TNamed named)
		throws Exception
	{
		if (named instanceof TOpaque) {
			return getRef((TOpaque) named);
		}
		else if (named instanceof TDef) {
			return getRef((TDef) named);
		}
		else if (named instanceof TParam) {
			return ((TParam)named).name;
		}
		else {
			throw badType(named);
		}
	}

	private String getRef(TDef def)
		throws Exception
	{
		if (def.value instanceof TIndirect) {
			return getRef((TIndirect) def.value);
		}
		else if (def.value instanceof TCompound) {
			return def.name;
		}
		else {
			throw badType(def.value);
		}
	}

	private String getRef(TOpaque base)
		throws Exception
	{
		if (base.equals(Type.LIST))    return "java.util.List";
		if (base.equals(Type.STRING))  return "java.lang.String";
		if (base.equals(Type.INTEGER)) return "java.math.BigInteger";
		if (base.equals(Type.BOOLEAN)) return "java.lang.Boolean";
		if (base.equals(Type.VOID))    return "java.lang.Void";
		if (base.equals(Type.MAYBE))   return "java.lang.Maybe";
		return null;
	}

	private String genParamDecl(Collection<TParam> params)
		throws Exception
	{
		if (params == null) return "";
		assert !params.isEmpty();
		StringBuilder buf = new StringBuilder();
		buf.append('<');
		for (TParam param : params) {
			if (!param.kind.equals(TKind.BASE)) throw new Exception(param, "can't handle higher-order type parameters");
			buf.append(param.name);
			buf.append(",");
		}
		buf.setLength(buf.length() - 1);
		buf.append('>');
		return buf.toString();
	}

  private void printHeader()
  {
    out.print("// Automatically Generated From: ");
    out.println(sourceFile);
    out.println();

    if (packageName != null) {
      out.print("package ");
			out.print(packageName);
			out.println(";");
    }
    out.println();
  }

	private RuntimeException badType(Object o)
	{
		return new RuntimeException("bad type: " + o.getClass().getName());
	}
}

