package cks.model.type;

import cks.model.*;

import java.io.PrintWriter;
import java.util.Map;

public abstract class TCompound extends Type
{
	public final Map<String,TMember> members;
	public final TMember implicit;
	public TIndirect parentRef;

	protected abstract void setParent(TCompound parentRef)
		throws ResolutionException;

	protected abstract TCompound getParent();
	
	protected abstract String getCategory();

	public TCompound(Location loc, Map<String,TMember> members, TMember implicit)
	{
		super(loc);
		this.members = members;
		this.implicit = implicit;

		assert implicit == null || members.values().contains(implicit);
	}

	protected void dumpChildren(PrintWriter out, int depth)
	{
		if (parentRef != null) {
			indent(out, depth);
			out.println("[parent]");
			parentRef.dump(out, depth + 1);
		}

		for (TMember member : members.values()) {
			indent(out, depth);

			out.print(member.name);
			if (member == implicit) out.print("*");
			out.println();

			member.type.dump(out, depth+1);
		}

		super.dumpChildren(out, depth);
	}

	// resolution -----------------------------------------------------

	protected void resolveNames_(MapStack<String,TNamed> context)
		throws ResolutionException
	{
		if (this.parentRef != null) {
			this.parentRef.resolveNames(context);
		}

		for (TMember member : members.values()) {
			member.type.resolveNames(context);
		}
	}

	protected void resolveKinds_()
		throws ResolutionException
	{
		resolveParent();

		// Make sure all field types are of kind '*'
		for (TMember member : this.members.values()) {
			member.type.calculateKind();
			TKind k = member.type.getKind();
			if (!k.equals(TKind.BASE)) {
				throw new ResolutionException(this, "members of a compound type must have a concrete type; "
						+ "found higher-order type [" + k + "]");
			}
		}

		for (TMember member : this.members.values()) {
			member.type.resolveKinds();
		}
	}

	private void resolveParent()
		throws ResolutionException
	{
		assert getParent() == null;

		if (this.parentRef != null) {
			this.parentRef.calculateKind();

			// Make sure the parent's kind is '*'
			if (!parentRef.getKind().equals(TKind.BASE)) {
				throw new ResolutionException(this, "cannot inherit from a type function");
			}

			// Make sure the parent class matches.
			TCompound realParent = this.parentRef.getBase();
			if (realParent == null || !realParent.getClass().equals(this.getClass())) {
				throw new ResolutionException(this, "cannot derive from \""
						+ parentRef.getIdent() + "\"; it isn't a " + getCategory() + " type");
			}

			setParent((TCompound) realParent);

			checkForConflicts();
		}
	}

	private void checkForConflicts()
		throws ResolutionException
	{
		TCompound ancestor = this;

		// Loop through the entire inheritance chain looking for duplicate
		// members.
		while (ancestor.getParent() != null) {

			String ancestorName = ancestor.parentRef.getIdent();
			ancestor = ancestor.getParent();

			// For each member in our ancestor's member list
			for (TMember ancestorMember : ancestor.members.values()) {

				if (this.members.containsKey(ancestorMember.name)) {
					// We found a dupe
					TMember myMember = this.members.get(ancestorMember.name);
					throw new ResolutionException(myMember,
							"member \"" + myMember.name + "\" already defined in ancestor \"" + ancestorName + "\"",
							ancestorMember);
				}

			}
		}
	}

	protected void calculateKind() {}
	public TKind getKind() { return TKind.BASE; }

	// properties -----------------------------------------------------

	protected TKind getKind_() { return TKind.BASE; }
	public final TCompound getBase() { return this; }
	//public void accept(TVisitor v) { v.visit(this); }
}
