package cks.model.type;

import cks.model.*;

import java.io.PrintWriter;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

public abstract class TDef extends Node implements TNamed
{
	public final String name;
	public final Type value;
	private TKind kind;

	protected TDef(Location loc, String name, Type value)
	{
		super(loc);

		assert name != null;
		assert value != null;

		this.name = name;
		this.value = value;
	}

	protected abstract void resolveNames(MapStack<String,TNamed> context)
		throws ResolutionException;

	protected void resolveKinds()
		throws ResolutionException
	{
		calculateKind();
		value.resolveKinds();
	}

	public void toString(StringBuilder b)
	{
		super.toString(b);
		b.append(": ");
		b.append(this.name);
	}

	protected void dumpChildren(PrintWriter out, int depth)
	{
		value.dump(out, depth);
		super.dumpChildren(out, depth);
	}

	public void calculateKind()
		throws ResolutionException
	{
		if (this.kind != null) return;
		this.value.calculateKind();
		this.kind = calculateKind(this.value.getKind());
	}

	protected abstract TKind calculateKind(TKind valueKind);

	public final TKind getKind()
	{
		assert this.kind != null;
		return this.kind;
	}

	public final TCompound getBase() { return this.value.getBase(); }

	public static class Simple extends TDef
	{
		public Simple(Location loc, String name, Type value)
		{
			super(loc, name, value);
		}

		protected void resolveNames(MapStack<String,TNamed> context)
			throws ResolutionException
		{
			value.resolveNames(context);
		}

		protected TKind calculateKind(TKind valueKind) { return valueKind; }
	}

	public static class Function extends TDef
	{
		public final Map<String,TParam> params;

		public Function(Location loc, String name, Map<String,TParam> params, Type value)
		{
			super(loc, name, value);

			assert params != null;

			this.params = params;
		}

		// resolve ----------------------------------------------------

		protected void resolveNames(MapStack<String,TNamed> context)
			throws ResolutionException
		{
			value.resolveNames(context.push(this.params));
		}

		protected TKind calculateKind(TKind valueKind) { return abs(this.params, valueKind); }

		private static TKind abs(Map<String,TParam> params, TKind base)
		{
			List<TKind> paramsK = new ArrayList<TKind>(params.size());
			for (TParam param : params.values()) {
				paramsK.add(param.getKind());
			}

			return base.abs(paramsK);
		}
		
		// print ----------------------------------------------------
		
		protected void dumpChildren(PrintWriter out, int depth)
		{
			for (TParam param : params.values()) {
				param.dump(out, depth);
			}
			super.dumpChildren(out, depth);
		}

	}
}
