package cks.model.type;

import cks.model.*;
import java.util.List;
import java.util.Iterator;
import java.io.PrintWriter;

public class TInvoke extends TIndirect
{
	public final TIndirect target;
	public final List<Type> args;
	public TKind kind;

	public TInvoke(Location location, TIndirect target, List<Type> args)
	{
		super(location);

		assert target != null;
		assert args != null;

		this.target = target;
		this.args = args;
	}

	// resolve ---------------------------------------------------

	protected void resolveNames_(MapStack<String,TNamed> context)
		throws ResolutionException
	{
		this.target.resolveNames(context);
		for (Type arg : args) {
			arg.resolveNames(context);
		}
	}

	protected void resolveKinds_()
		throws ResolutionException
	{
		this.target.resolveKinds();
		for (Type arg : args) {
			arg.resolveKinds();
		}
	}

	protected void calculateKind()
		throws ResolutionException
	{
		if (this.kind != null) return;

		this.target.calculateKind();

		// Make sure the target is a function.
		TKind targetKind = this.target.getKind();
		if (targetKind instanceof TKind.Base) {
			throw new ResolutionException(this, "attempt to invoke a type that isn't a type function; "
					+ "its kind is [" + targetKind + "]");
		}

		TKind.Function funcKind = (TKind.Function) targetKind;
		List<TKind> expectedKinds = funcKind.params;

		// Check if the function signature matches.  We could just call
		// TKind.equals(...), but we want a nicer error message.

		// Make sure we have the correct number of arguments
		if (expectedKinds.size() != this.args.size()) {
			throw new ResolutionException(this, "incorrect number of arguments in invocation of \""
					+ target.getIdent() + "\"; expecting " + expectedKinds.size() + " found "
					+ this.args.size());
		}

		// Now check each argument to make sure the kinds are exactly equal.
		Iterator<TKind> ei = expectedKinds.iterator();
		Iterator<Type> ai = this.args.iterator();
		int argPos = 0;
		while (ei.hasNext()) {
			argPos++;
			assert ai.hasNext();
			TKind expected = ei.next();
			Type arg = ai.next();
			arg.calculateKind();
			TKind found = arg.getKind();
			if (!expected.equals(found)) {
				throw new ResolutionException(this, "kind mismatch in argument " + argPos
						+ " of invocation; expected [" + expected + "], found [" + found + "]");
			}
		}

		this.kind = funcKind.result;
		assert this.kind != null;
	} 

	// properties ---------------------------------------------------

	public TKind getKind()
	{
		assert kind != null;
		return this.kind;
	}

	public TCompound getBase() { return this.target.getBase(); }
	public String getIdent() { return this.target.getIdent() + "(...)"; }
	//public void accept(TypeVisitor v) { v.visit(this); }

	// dump ---------------------------------------------------

	protected void dumpChildren(PrintWriter out, int depth)
	{
		target.dump(out, depth);
		for (Type arg : this.args) {
			arg.dump(out, depth);
		}
		super.dumpChildren(out, depth);
	}

}
