package cks.parser;

import cks.model.Location;

import java.io.IOException;
import java.util.Stack;

public final class LayoutTokenManager
{
  private TokenSource source;
  private Token next; // lookahead
  private int depth;

	private Token pending;
	private String pendingIndent;
	private Stack<String> indentLevels;

  public LayoutTokenManager(TokenSource source)
  {
    this.source = source;
    this.next = null;
    this.depth = 0;

		this.indentLevels = new Stack<String>();
		this.indentLevels.push("");
  }

  // take
  public Token take()
    throws SyntaxException, IOException
  {
    Token token;
    if (next == null) {
      token = get();
    } else {
      token = next;
      next = null;
    }

    updateDepth(token);
    return token;
  }

  // peek
  public Token peek()
    throws SyntaxException, IOException
  {
    if (next == null) {
      next = get();
    }
    return next;
  }

  // next
  public void next()
    throws SyntaxException, IOException
  {
    if (next == null) {
      throw new RuntimeException("you're not allowed to ignore tokens, punk");
    }

    updateDepth(next);
    next = null;
  }

  // updateDepth
  private void updateDepth(Token token)
  {
    if (token.type == Codes.Indent) {
      depth++;
    } else if (token.type == Codes.Dedent) {
      depth--;
    }
  }

  // depth
  public int depth()
  {
    return depth;
  }

  // unwind
  public void unwind(int startDepth)
    throws SyntaxException, IOException
  {
    if (peek().type == Codes.EOL) next();
    while (depth() > startDepth) {
      expect(take(), Codes.Dedent);
    }
  }

  // unwind (endMark)
  public void unwind(int startDepth, int endMark)
    throws SyntaxException, IOException
  {
    while (true) {
      Token token = take();

      if (token.type == Codes.Dedent) {
        if (depth() == startDepth) {
          expect(take(), endMark);
          break;
        }
      }
      else if (token.type == Codes.EOL) {
        // ignore
      }
      else if (token.type == endMark) {
        break;
      }
      else {
        expect(token, endMark);
      }
    }
  }

  // expect(token)
  public void expect(Token token, int type)
    throws SyntaxException
  {
    if (token.type != type) {
      StringBuilder error = new StringBuilder("expecting ");
      Token.toStringContent(error, type);
      throw new SyntaxException(token, error.toString());
    }
  }

  // expect
  public void expect(int type)
    throws SyntaxException, IOException
  {
    expect(take(), type);
  }

	private Token get()
		throws SyntaxException, IOException
	{
		// We don't have an indent/dedent backlog.
		if (pending == null) {

			Token token = source.get();

			// If we got an EOL, then we're going to have some indent/dedent backlog
			if (token.type == Codes.EOL) {
				
				Token next;
				String pendingIndent = "";

				do {
					next = source.get();
				} while (next.type != Codes.EOL); // skip over repeated EOLs

				pending = next;
				pendingIndent = source.getIndent();

				String top = indentLevels.peek();
				if (pendingIndent.equals(top)) {
					// No indent/dedent required, return the EOL
					return token;
				}
				else if (top.startsWith(pendingIndent)) {
					// Indent
					indentLevels.push(pendingIndent);
					return mkFake(Codes.Indent, next.loc);
				}
				else if (pendingIndent.startsWith(top)) {
					// Dedent
					indentLevels.pop();
					return mkFake(Codes.Dedent, next.loc);
				}
				else {
					// Layout mismatch
					throw new SyntaxException(next, "tab/space mismatch in indentation");
				}

			}
			// If it's not an EOL, then just return it.
			else {
				return token;
			}
		}

		// We do have extra indent/dedent work to do.
		else {
			String top = indentLevels.peek();
			if (pendingIndent.equals(top)) {
				// We've hit the right level, remove the backlog token and return
				// it.
				Token temp = pending;
				pendingIndent = null;
				pending = null;
				return temp;
			}
			else if (top.startsWith(pendingIndent)) {
				// Still more dedents to go...
				indentLevels.pop();
				return mkFake(Codes.Dedent, pending.loc);
			}
			else {
				// Bad indentation level.
				throw new SyntaxException(pending, "bad indentation level");
			}
		}

	}

	private Token mkFake(int code, Location loc)
	{
		Token token = new Token(code);
		token.text = "";
		token.loc = loc;
		return token;
	}


}
