package cks.parser;

import cks.model.*;
import cks.model.type.*;

import java.util.Map;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

import java.io.IOException;
import java.io.FileReader;
import java.io.PrintWriter;

public final class SimpleTypeParser
{
	private SimpleTypeParser(SimpleTokenManager tm)
	{
		this.tm = tm;
	}

	public static Type parse(TokenSource ts)
		throws SyntaxException, IOException
	{
		return parse(new SimpleTokenManager(ts));
	}

	public static Type parse(SimpleTokenManager ts)
		throws SyntaxException, IOException
	{
		SimpleTypeParser parser = new SimpleTypeParser(ts);
		Type type = parser.parse();
		return type;
	}

	private SimpleTokenManager tm;

	private Type parse()
		throws SyntaxException, IOException
	{
		Type type = pTopLevel();
		expect(Codes.EOF);
		return type;
	}

	private Type pTopLevel()
		throws SyntaxException, IOException
	{
		return pType();
	}

	private Type pType()
		throws SyntaxException, IOException
	{
		Type type;

		Token token = peek();
		switch (token.type) {

			case '[': {
				type = pList();
				break;
			}

			case '<': {
				type = pVariant();
				break;
			}

			case '{': {
				type = pRecord();
				break;
			}

			case '(': {
				next();
				expect(')');
				type = new TOpaqueRef(token.loc, Type.VOID);
				break;
			}

			case Codes.Ident: {
				type = pRef();
				break;
			}

			default:
				throw new SyntaxException(token, "expecting a type");
		}

		// The "maybe" type.
		while (peek().type == '?') {
			Token t = take();
			type = new TInvoke(t.loc,
					new TOpaqueRef(t.loc, Type.MAYBE),
					Collections.singletonList(type));
		}

		return type;
	}

	private Type pList()
		throws SyntaxException, IOException
	{
		Token first = take();
		assert first.type == '[';

		Type elementType = pType();

		expect(']');

		return new TInvoke(first.loc,
				new TOpaqueRef(first.loc, Type.LIST),
				Collections.singletonList(elementType));
	}

	private TRecord pRecord()
		throws SyntaxException, IOException
	{
		return pCompound('{', '}', ',', CompoundFactory.RECORD, "record");
	}

	private TVariant pVariant()
		throws SyntaxException, IOException
	{
		return pCompound('<', '>', '|', CompoundFactory.VARIANT, "variant");
	}

	private void pDef(Token deftok, Map<String,TDef> defs)
		throws SyntaxException, IOException
	{
		assert deftok.type == Codes.Ident ;
		assert deftok.text.equals("def") ;

		Token name = take();
		assert name.type == Codes.Ident;

		Map<String,TParam> params = pDeclareParams();

		if (peek().type == '{' || peek().type == '<') {
			// Don't need an '='
		} else {
			expect('=');
		}

		Type value = pType();

		TDef def;
		if (params != null) {
			def = new TDef.Function(deftok.loc, name.text, params, value);
		} else {
			def = new TDef.Simple(deftok.loc, name.text, value);
		}

		TDef displaced = defs.put(name.text, def);
		if (displaced != null) {
			throw new SyntaxException(name.loc,
					"duplicate type definition \"" + name.text + "\"; original definition", displaced.loc);
		}
	}

	private Map<String,TParam> pDeclareParams()
		throws SyntaxException, IOException
	{
		if (peek().type != '(') return null;
		next();

		Map<String,TParam> params = new LinkedHashMap<String,TParam>();

		if (peek().type == ',') next();

		while (peek().type != ')') {
			Token name = expect(Codes.Ident);

			// TODO: higher-kinded type parameters

			TParam newGuy = new TParam(name.loc, name.text, TKind.BASE);
			TParam oldGuy = params.put(name.text, newGuy);

			if (oldGuy != null) {
				throw new SyntaxException(name.loc,
						"duplicate parameter \"" + name.text + "\"; original parameter",
						oldGuy.loc);
			}

			if (peek().type == ')') break;
			expect(',');
		}
		next();

		return params;
	}

	private static abstract class CompoundFactory<T extends TCompound> {
		public abstract T make(Location loc, Map<String,TMember> members, TMember implicit);

		public static final CompoundFactory<TRecord> RECORD = new CompoundFactory<TRecord>() {
			public TRecord make(Location loc, Map<String,TMember> members, TMember implicit)
			{ return new TRecord(loc, members, implicit); }
		};
		public static final CompoundFactory<TVariant> VARIANT = new CompoundFactory<TVariant>() {
			public TVariant make(Location loc, Map<String,TMember> members, TMember implicit)
			{ return new TVariant(loc, members, implicit); }
		};
	}

	private <T extends TCompound>
	T pCompound(int startMark, int endMark, int divMark, CompoundFactory<? extends T> factory, String kindName)
		throws SyntaxException, IOException
	{
		Token first = take();
		assert first.type == startMark;

		Map<String,TMember> members = new LinkedHashMap<String,TMember>();
		TMember implicit = null; // The "implicit" member (for mixed content)
		Map<String,TDef> defs = new LinkedHashMap<String,TDef>();

		boolean needDivider = false;

		// The list of div-separated members.
		while (peek().type != endMark) {
			if (needDivider && !tm.prevWasEol()) {
				expect(',');
			}

			Token name = expect(Codes.Ident);

			// Named type definition
			if (name.text.equals("def") && peek().type == Codes.Ident) {
				pDef(name, defs);
			}
			// Member entry
			else {
				Type type;
				boolean thisMemberIsImplicit = false;

				// Marker for the "implicit" member.
				if (peek().type == '*') {
					if (implicit != null) {
						throw new SyntaxException(name.loc,
							"extra implicit member marker on \"" + name.text + "\"; first marker", implicit.loc);
					}
					next();
					thisMemberIsImplicit = true;
				}

				if (peek().type == ':') {
					// If there's a ':' then parse the type.
					next();
					type = pType();
				} else if (peek().type == '<' || peek().type == '{') {
					// The ':' is optional if the type is a compound type.  Even though
					// we're calling pType(), we've already checked that it's a compound
					// type.
					type = pType();
				} else {
					// Defaults to TVoid
					type = new TOpaqueRef(null, Type.VOID);
				}

				TMember newMember = new TMember(name.loc, name.text, type);
				TMember displaced = members.put(name.text, newMember);
				if (displaced != null) {
					throw new SyntaxException(name.loc,
							"duplicate member \"" + name.text + "\"; original definition", displaced.loc);
				}

				if (thisMemberIsImplicit) implicit = newMember;
			}

			if (peek().type == ',') {
				next();
				needDivider = false;
			} else {
				needDivider = true;
			}
		}

		next();
		
		T ret = factory.make(first.loc, members, implicit);
		ret.defs = defs;
		return ret;
	}

	private Type pRef()
		throws SyntaxException, IOException
	{
		Token ident = take();
		assert(ident.type == Codes.Ident);

		TIndirect ref = new TRef(ident.loc, ident.text);

		// invocation parameter lists
		while (true) {
			List<Type> params = pInvokeParams();
			if (params == null) break;
			ref = new TInvoke(ident.loc, ref, params);
		}

		if (peek().type == '+') {
			next();

			Type type = pType();
			if (type instanceof TCompound) {
				TCompound combine = (TCompound) type;
				combine.parentRef = ref;
				return combine;
			}

			throw new SyntaxException(type.loc, "only records and variants can inherit");
		}
		else {
			return ref;
		}
	}

	private List<Type> pInvokeParams()
		throws SyntaxException, IOException
	{
		if (peek().type != '(') return null;
		next();

		List<Type> params = new ArrayList<Type>();

		if (peek().type == ',') next();

		while (peek().type != ')') {
			params.add(pType());
			if (peek().type == ')') break;
			expect(',');
		}
		next();

		return params;
	}

	private Token take()
		throws IOException, SyntaxException
	{
		Token token = tm.take();
		return token;
	}

	private Token peek()
		throws IOException, SyntaxException
	{
		Token token = tm.peek();
		return token;
	}

	private void next()
		throws IOException, SyntaxException
	{
		tm.next();
	}

	private Token expect(int type)
		throws IOException, SyntaxException
	{
		return tm.expect(type);
	}

	private Type tokenDump()
		throws SyntaxException, IOException
	{
		while (true) {
			Token token = tm.take();
			System.out.println("# " + token);
			if (token.type == Codes.EOF) break;
		}
		return null;
	}

}
