package cks.parser;

import cks.model.*;
import cks.model.value.*;
import cks.model.type.*;

import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Stack;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;

import java.io.IOException;
import java.io.FileReader;
import java.io.PrintWriter;

public final class SimpleValueParser
{
	private SimpleTokenManager tm;

	private SimpleValueParser(SimpleTokenManager tm)
	{
		this.tm = tm;
	}

	enum Mode { Normal, Tag, Markup, }

	public static Value parse(TokenSource ts, Type type)
		throws SyntaxException, IOException
	{
		return parse(new SimpleTokenManager(ts), type);
	}

	public static Value parse(SimpleTokenManager ts, Type type)
		throws SyntaxException, IOException
	{
		SimpleValueParser parser = new SimpleValueParser(ts);
		Value value = parser.parse(type);
		return value;
	}

	private Value parse(Type type)
		throws SyntaxException, IOException
	{
		Value value = pValue(null, type, Mode.Normal);
		expect(Codes.EOF);
		return value;
	}

	private Value pValue(Type type, Mode mode)
		throws SyntaxException, IOException
	{
		return pValue(null, type, mode);
	}

	private Value pValue(String name, Type type, Mode mode)
		throws SyntaxException, IOException
	{
		assert type != null;

		if (type instanceof TRecord) {
			return pRecord(name, (TRecord) type, mode);
		} else if (type instanceof TVariant) {
			return pVariant(name, (TVariant) type, mode);
		} else if (type instanceof TInvoke) {
			return pInvoke(name, (TInvoke) type, mode);
		} else if (type instanceof TOpaqueRef) {
			return pOpaque(name, ((TOpaqueRef) type).target, mode);
		} else if (type instanceof TRef) {
			return pRef(name, (TRef) type, mode);
		} else {
			throw badType(type);
		}
	}

	// I should be passing these around as parameters, but I got lazy.

	// Bindings from type parameters to actual types.  Mappings
	// are added by TDef.Function and looked up by TParam
	private MapStack<TParam,Type> paramBindings = MapStack.empty();

	// List of type arguments waiting to be bound to type parameters.
	// Parameter lists are added by TInvoke and consumed by TDef.Function
	// and TOpaque.
	private Stack<List<Type>> pendingTypeArgs = new Stack<List<Type>>();

	// When in markup mode, push the name of the open tag onto the stack
	// so that we can check for the correct close tag.
	private Stack<Token> openTagStack = new Stack<Token>();

	private Value pRef(String name, TRef type, Mode mode)
		throws SyntaxException, IOException
	{
		TNamed target = type.target;
		if (target instanceof TOpaque) {
			return pOpaque(name, (TOpaque) target, mode);
		}
		else if (target instanceof TDef) {
			return pDef(name, (TDef) target, mode);
		}
		else if (target instanceof TParam) {
			return pParam(name, (TParam) target, mode);
		}
		else {
			throw badType(target);
		}
	}

	private Value pDef(String name, TDef type, Mode mode)
		throws SyntaxException, IOException
	{
		if (type instanceof TDef.Function) {
			TDef.Function func = (TDef.Function) type;
			assert pendingTypeArgs.size() > 0;

			List<Type> params = pendingTypeArgs.pop();
			assert params.size() == func.params.size();

			// Store the (parameter -> argument) relationships
			Map<TParam,Type> map = new HashMap<TParam,Type>();
			Iterator<TParam> formal = func.params.values().iterator();
			Iterator<Type> actual = params.iterator();
			while (formal.hasNext()) {
				assert actual.hasNext();
				map.put(formal.next(), actual.next());
			}

			MapStack<TParam,Type> old = this.paramBindings;
			this.paramBindings = old.push(map);
			Value v = pValue(type.name, type.value, mode);
			this.paramBindings = old;

			return v;
		}
		else if (type instanceof TDef.Simple) {
			return pValue(type.name, type.value, mode);
		}
		else {
			throw badType(type);
		}
	}

	private Value pInvoke(String name, TInvoke invoke, Mode mode)
		throws SyntaxException, IOException
	{
		TIndirect target = invoke.target;
		pendingTypeArgs.push(invoke.args);
		return pIndirect(name, target, mode);
	}

	private Value pIndirect(String name, TIndirect type, Mode mode)
		throws SyntaxException, IOException
	{
		assert type != null;

		if (type instanceof TRef) {
			return pRef(name, (TRef) type, mode);
		}
		else if (type instanceof TInvoke) {
			return pInvoke(name, (TInvoke) type, mode);
		}
		else if (type instanceof TOpaqueRef) {
			return pOpaque(name, ((TOpaqueRef) type).target, mode);
		}
		else {
			throw badType(type);
		}
	}

	private Value pParam(String name, TParam param, Mode mode)
		throws SyntaxException, IOException
	{
		Type bound = paramBindings.get(param);
		if (bound == null) throw die("unbound parameter type: " + param);
		return pValue(bound, mode);
	}

	private Value pOpaque(String name, TOpaque type, Mode mode)
		throws SyntaxException, IOException
	{
		switch (mode) {

			case Normal:
				return pOpaqueNormal(type);

			case Tag:
				if (peek().type == '>') {
					next();
					Value v = pOpaqueMarkup(name, type);
					expectCloseTag();
					return v;
				} else {
					Value v = pOpaqueNormal(type);
					expectEmptyTag();
					return v;
				}

			case Markup:
				return pOpaqueMarkup(name, type);
		}

		throw badType(mode);
	}

	private Value pOpaqueMarkup(String name, TOpaque type)
		throws SyntaxException, IOException
	{
		if (type == Type.STRING) {
			// Take a string token
			assert pendingTypeArgs.size() == 0;
			// TODO: switch lexer to markup mode and all characters up.  For now,
			// we're just eating up idents, ints, and string.
			StringBuilder b = new StringBuilder();
			Token first = peek();
			Token t = first;
			while (t.type == Codes.Ident || t.type == Codes.LitString || t.type == Codes.LitInt) {
				b.append(t.text);
				b.append(' ');
				next();
				t = peek();
			}
			if (b.length() > 0) {
				b.setLength(b.length() - 1);
			}
			return new VPrimitive.String(first.loc, b.toString());
		}
		else if (type == Type.INTEGER) {
			// Take an integer token
			assert pendingTypeArgs.size() == 0;
			Token.LitInt tok = (Token.LitInt) expect(Codes.LitInt);
			return new VPrimitive.Integer(tok.loc, tok.value);
		}
		else if (type == Type.BOOLEAN) {
			// Parse with our fake BOOLEAN_EQUIV type.
			assert pendingTypeArgs.size() == 0;
			VVariant equivValue = pVariant(null, BOOLEAN_EQUIV, Mode.Markup);
			boolean value;
			if (equivValue.label == BOOLEAN_EQUIV_TRUE.name) {
				value = true;
			}
			else if (equivValue.label == BOOLEAN_EQUIV_FALSE.name) {
				value = false;
			}
			else {
				throw die("no match for label: " + equivValue.label);
			}
			return new VPrimitive.Void(null);
		}
		else if (type == Type.VOID) {
			// Nothing to do!
			assert pendingTypeArgs.size() == 0;
			return new VPrimitive.Void(null);
		}
		else if (type == Type.LIST) {
			// Nothing to do!
			assert pendingTypeArgs.size() == 1;
			List<Type> params = pendingTypeArgs.pop();
			assert params.size() == 1;
			return pList(null, params.get(0), Mode.Markup);
		}
		else {
			throw new RuntimeException("Can't parse opaque type: " + type);
		}
	}

	private Value pOpaqueNormal(TOpaque type)
		throws SyntaxException, IOException
	{
		if (type == Type.STRING) {
			// Take a string token
			assert pendingTypeArgs.size() == 0;
			Token.LitStr tok = (Token.LitStr) expect(Codes.LitString);
			return new VPrimitive.String(tok.loc, tok.value);
		}
		else if (type == Type.INTEGER) {
			// Take an integer token
			assert pendingTypeArgs.size() == 0;
			Token.LitInt tok = (Token.LitInt) expect(Codes.LitInt);
			return new VPrimitive.Integer(tok.loc, tok.value);
		}
		else if (type == Type.BOOLEAN) {
			// Parse with our fake BOOLEAN_EQUIV type.
			assert pendingTypeArgs.size() == 0;
			VVariant equivValue = pVariant(null, BOOLEAN_EQUIV, Mode.Normal);
			boolean value;
			if (equivValue.label == BOOLEAN_EQUIV_TRUE.name) {
				value = true;
			}
			else if (equivValue.label == BOOLEAN_EQUIV_FALSE.name) {
				value = false;
			}
			else {
				throw die("no match for label: " + equivValue.label);
			}
			return new VPrimitive.Void(null);
		}
		else if (type == Type.VOID) {
			// Nothing to do!
			assert pendingTypeArgs.size() == 0;
			return new VPrimitive.Void(null);
		}
		else if (type == Type.LIST) {
			// Nothing to do!
			assert pendingTypeArgs.size() == 1;
			List<Type> params = pendingTypeArgs.pop();
			assert params.size() == 1;
			return pList(null, params.get(0), Mode.Normal);
		}
		else {
			throw new RuntimeException("Can't parse opaque type: " + type);
		}
	}

	private VList pList(String name, Type type, Mode mode)
		throws SyntaxException, IOException
	{
		switch (mode) {

			case Normal:
				return pListNormal(type);

			case Tag:
				Token t = peek();
				if (t.type == '>') {
					next();
					VList l = pListMarkup(t.loc, type);
					expectCloseTag();
					return l;
				} else if (t.type == Codes.LeafTagEnd) {
					next();
					openTagStack.pop();
					return new VList(t.loc, Collections.<Value>emptyList());
				} else {
					VList l = pListNormal(type);
					expectEmptyTag();
					return l;
				}

			case Markup:
				return pListMarkup(null, type);

			default:
				throw badType(mode);
		}
	}

	private VList pListNormal(Type type)
		throws SyntaxException, IOException
	{
		Token start = expect('[');
		List<Value> elements = new ArrayList<Value>();

		boolean needDivider = false;

		while (peek().type != ']') {
			if (needDivider && !tm.prevWasEol()) {
				expect(',');
			}

			elements.add(pValue(null, type, Mode.Normal));

			if (peek().type == ',') {
				next();
				needDivider = false;
			} else {
				needDivider = true;
			}
		}

		expect(']');
		return new VList(start.loc, elements);
	}

	private VList pListMarkup(Location start, Type type)
		throws SyntaxException, IOException
	{
		List<Value> elements = new ArrayList<Value>();

		while (true) {
			if (peek().type == Codes.CloseTagStart) break;

			Value value = pValue(null, type, Mode.Markup);
			elements.add(value);
		}

		// If the starting location wasn't passed in, look for it elsewhere.
		if (start == null) {
			if (elements.size() > 0) {
				// Take the location of the first element.
				start = elements.get(0).loc;
			}
			else {
				// Take the location of the closing tag.
				start = peek().loc;
			}
		}

		return new VList(start, elements);
	}

	private Value pRecord(String name, TRecord type, Mode mode)
		throws SyntaxException, IOException
	{
		switch (mode) {

			case Normal:
				Token token = take();
				switch (token.type) {

					// direct syntax
					case '{': return pRecordNormal(token.loc, type);

					// markup syntax
					case '<': return pRecordTag(name, type, token);

					default:
						throw new SyntaxException(token, "expecting a record");
				}

			case Tag:
				return pRecordTagContinue(type);

			case Markup:
				return pRecordMarkup(null, type);

		}

		throw badType(mode);
	}

	private VRecord pRecordNormal(Location loc, TRecord type)
		throws SyntaxException, IOException
	{
		Map<String,VField> fields = new LinkedHashMap<String,VField>();
		pRecordFields(type, fields);
		expect('}');
		VRecord value = mkRecord(loc, type, fields);
		return value;
	}

	private void pRecordFields(TRecord type, Map<String,VField> fields)
		throws SyntaxException, IOException
	{
		boolean needDivider = false;

		while (true) {

			if (peek().type == Codes.Ident) {
				if (needDivider && !tm.prevWasEol()) {
					throw new SyntaxException(peek(), "fields on the same line need to be separated by commas");
				}
			} else if (peek().type == ',') {
				take();
			} else {
				break;
			}

			// field name
			Token labelToken = expect(Codes.Ident);
			String label = labelToken.text;

			// check if the field exists
			TMember field = type.members.get(label);
			if (field == null) {
				throw new SyntaxException(labelToken, "record has no field named \"" + label + "\"");
			}

			if (peek().type == '{') {
				// don't need an '='
			} else {
				expect('=');
			}

			// field value
			Value value = pValue(field.type, Mode.Normal);

			// make sure this field hasn't been defined already
			addField(fields, labelToken.loc, field.name, value);

			if (peek().type == ',') {
				next();
				needDivider = false;
			} else {
				needDivider = true;
			}
		}
	}

	private VRecord pRecordTag(String name, TRecord type, Token openTagBracket)
		throws SyntaxException, IOException
	{
		// 'tagOpen' saves the token of the tag's label.  If there is no label
		// then we'll default to the token of the opening angle bracket.
		Token tagOpen = openTagBracket;

		// "<>"
		if (peek().type == '>') {
		}
		// "<. [...]>"
		else if (peek().type == '.') {
			next();
		}
		// "<BlahBlah [...]>"
		else if (peek().type == Codes.Ident) {
			Token tagLabel = take();
			if (name == null) {
				throw new SyntaxException(tagLabel, "the tag name \"" + tagLabel.text + "\" cannot be accepted because the type has no name");
			}
			if (!tagLabel.text.equals(name)) {
				throw new SyntaxException(tagLabel, "the tag name \"" + tagLabel.text + "\" doesn't match the type name \"" + name + "\"");
			}
			tagOpen = tagLabel;
		}

		openTag(tagOpen);

		VRecord value = pRecordTagContinue(type);

		return value;
	}

	boolean tagIsOpen = false;

	private void openTag(Token t)
	{
		openTagStack.push(t);
	}

	private void expectEmptyTag()
		throws SyntaxException, IOException
	{
		// If it's a leaf tag, then our job is done.
		if (peek().type == Codes.LeafTagEnd) {
			next();
			openTagStack.pop();
			return;
		}
		
		expect('>');
		expectCloseTag();
	}

	private void expectCloseTag()
		throws SyntaxException, IOException
	{
		Token t = take();
		assert !openTagStack.isEmpty() : t.loc;
		Token openTag = openTagStack.pop();

		if (t.type != Codes.CloseTagStart) {
			throw new SyntaxException(t.loc, "expecting closing tag for tag opened at", openTag.loc);
		}

		if (peek().type == Codes.Ident) {
			Token closeTagIdent = take();
			if (openTag.type != Codes.Ident) {
				throw new SyntaxException(closeTagIdent.loc, "closing tag has label but corresponding opening tag does not");
			}
			if (!closeTagIdent.text.equals(openTag.text)) {
				throw new SyntaxException(closeTagIdent.loc, "tag label mismatch, opened with \"" + openTag.text + "\", closed with \"" + closeTagIdent.text + "\"; opening tag", openTag.loc);
			}
		}
		expect('>');
	}

	private VRecord pRecordTagContinue(TRecord type)
		throws SyntaxException, IOException
	{
		Map<String,VField> fields = new LinkedHashMap<String,VField>();

		pRecordFields(type, fields);

		Location start;
		if (fields.size() > 0) {
			start = fields.values().iterator().next().loc;
		} else {
			start = peek().loc;
		}

		if (peek().type == Codes.LeafTagEnd) {
			next();
			openTagStack.pop();
			return mkRecord(start, type, fields);
		}

		expect('>');

		VRecord v = pRecordMarkup(start, type, fields);

		expectCloseTag();

		return v;
	}

	private VRecord pRecordMarkup(Location start, TRecord type)
		throws SyntaxException, IOException
	{
		Map<String,VField> fields = new LinkedHashMap<String,VField>();
		return pRecordMarkup(start, type, fields);
	}

	private VRecord pRecordMarkup(Location start, TRecord type, Map<String,VField> fields)
		throws SyntaxException, IOException
	{
		// All content goes into the implicit field.
		if (type.implicit != null) {
			Value value = pValue(null, type.implicit.type, Mode.Markup);
			addField(fields, value.loc, type.implicit.name, value);
		}
		// The content contains multiple field entries.
		else {
			while (peek().type == '<') {
				next();
				if (peek().type == Codes.CloseTagStart) break;

				// field name
				Token labelToken = expect(Codes.Ident);
				String label = labelToken.text;
				openTag(labelToken);

				// check if the field exists
				TMember field = type.members.get(label);
				if (field == null) {
					throw new SyntaxException(labelToken, "record has no field named \"" + label + "\"");
				}

				Value value = pValue(null, field.type, Mode.Tag);

				addField(fields, labelToken.loc, field.name, value);
			}
		}

		// If the starting location wasn't passed in, look for it elsewhere.
		if (start == null) {
			if (fields.size() > 0) {
				// Take the location of the first field.
				start = fields.values().iterator().next().loc;
			}
			else {
				// Take the location of the closing tag.
				start = peek().loc;
			}
		}

		return mkRecord(start, type, fields);
	}

	private void addField(Map<String,VField> fields, Location loc, String name, Value value)
		throws SyntaxException
	{
		VField displaced = fields.put(name, new VField(loc, name, value));
		if (displaced != null) {
			throw new SyntaxException(loc,
					"duplicate definition of field \"" + name + "\"; original definition", displaced.loc);
		}
	}

	private VVariant pVariant(String name, TVariant type, Mode mode)
		throws SyntaxException, IOException
	{
		switch (mode) {

			case Normal:
				return pVariantNormal(type);

			case Markup:
				return pVariantMarkup(type);

			case Tag:
				return pVariantTagContinue(type);
		}

		throw badType(mode);
	}

	private VVariant pVariantNormal(TVariant type)
		throws SyntaxException, IOException
	{
		Token labelToken = expect(Codes.Ident);
		String label = labelToken.text;
		TMember selected = type.members.get(label);
		if (selected == null) {
			throw new SyntaxException(labelToken, "variant has no option named \"" + label + "\"");
		}
		label = selected.name; // Use the same string object (saves a little space)

		Value value = pValue(selected.type, Mode.Normal);
		return new VVariant(labelToken.loc, label, value);
	}

	private VVariant pVariantMarkup(TVariant type)
		throws SyntaxException, IOException
	{
		if (type.implicit != null && peek().type == Codes.Ident) {
			Token start = take();
			String s = start.text;
			while (peek().type == Codes.Ident) {
				s += " " + take().text;
			}
			return new VVariant(start.loc, type.implicit.name,
					new VPrimitive.String(start.loc, s));
		}
		else {
			expect('<');
			
			Token labelToken = expect(Codes.Ident);
			openTag(labelToken);

			return pVariantTagContinue(type, labelToken);
		}
	}

	private VVariant pVariantTagContinue(TVariant type)
		throws SyntaxException, IOException
	{
		if (peek().type == '>') {
			next();
			VVariant v = pVariant(null, type, Mode.Markup);
			expectCloseTag();
			return v;
		} else {
			Token labelToken = expect(Codes.Ident);
			return pVariantTagContinue(type, labelToken);
		}
	}

	private VVariant pVariantTagContinue(TVariant type, Token labelToken)
		throws SyntaxException, IOException
	{
		assert labelToken.type == Codes.Ident;

		String label = labelToken.text;
		TMember selected = type.members.get(label);
		if (selected == null) {
			throw new SyntaxException(labelToken, "variant has no option named \"" + label + "\" (valid options: " + type.members.keySet() + ")");
		}

		Value v;
		if (peek().type == '>') {
			next();
			v = pValue(selected.type, Mode.Markup);
			expectCloseTag();
		}
		else {
			v = pValue(selected.type, Mode.Tag);
		}

		return new VVariant(labelToken.loc, label, v);
	}

	// ------------------------------------------------
	// Type objects used only to represent the syntax of an opaque type.

	private static final TMember BOOLEAN_EQUIV_TRUE = mkMember("True");
	private static final TMember BOOLEAN_EQUIV_FALSE = mkMember("False");
	private static final TVariant BOOLEAN_EQUIV =
		mkVariant(BOOLEAN_EQUIV_TRUE, BOOLEAN_EQUIV_FALSE);

	private static final TMember mkMember(String name)
	{
		return new TMember(null, name, new TOpaqueRef(null, Type.VOID));
	}

	private static final TVariant mkVariant(TMember... options)
	{
		Map<String,TMember> optionMap = new HashMap<String,TMember>();
		for (TMember option : options) {
			TMember displaced = optionMap.put(option.name, option);
			assert displaced == null;
		}

		TVariant v = new TVariant(null, optionMap, null);

		try {
			v.resolve(Type.EMPTY_CONTEXT);
		} catch (ResolutionException ex) {
			throw new AssertionError(ex);
		}

		return v;
	}

	// ------------------------------------------------

	private VRecord mkRecord(Location loc, TRecord type, Map<String,VField> fields)
		throws SyntaxException, IOException
	{
		// make sure all the required fields have been defined
		for (TMember field : type.members.values()) {
			if (!fields.containsKey(field.name)) {
				Value defaultValue = mkDefaultValue(field.type);
				if (defaultValue == null) {
					System.out.println("field.type = " + field.type);
					throw new SyntaxException(loc, "missing record field \"" + field.name + "\" (and no default is available)");
				}
			}
		}

		return new VRecord(loc, fields);
	}

	private Value mkDefaultValue(Type type)
		throws SyntaxException, IOException
	{
		if (type instanceof TOpaqueRef) {
			// Void
			TOpaque t = ((TOpaqueRef) type).target;
			if (t.equals(Type.VOID)) {
				return new VPrimitive.Void(null);
			}
		}
		else if (type instanceof TInvoke) {
			TInvoke inv = (TInvoke) type;
			if (inv.target instanceof TOpaqueRef) {
				// Maybe
				TOpaque t = ((TOpaqueRef) inv.target).target;
				if (t.equals(Type.MAYBE)) {
					return new VPrimitive.Maybe(null, null);
				}
			}
		}

		return null;
	}

	// ------------------------------------------------

	private Token take()
		throws IOException, SyntaxException
	{
		Token token = tm.take();
//		System.err.println("TAKE: " + token);
		return token;
	}

	private Token peek()
		throws IOException, SyntaxException
	{
		Token token = tm.peek();
 //	 System.err.println("PEEK: " + token);
		return token;
	}

	private void next()
		throws IOException, SyntaxException
	{
	//	System.err.println("NEXT");
		tm.next();
	}

	private Token expect(int type)
		throws IOException, SyntaxException
	{
	//	System.err.println("NEXT");
		return tm.expect(type);
	}

	private Value tokenDump()
		throws SyntaxException, IOException
	{
		while (true) {
			Token token = tm.take();
			System.out.println("# " + token);
			if (token.type == Codes.EOF) break;
		}
		return null;
	}

	private AssertionError badType(Object o)
	{
		return new AssertionError("bad type: " + o.getClass().getName());
	}

	private AssertionError die(String s)
	{
		return new AssertionError(s);
	}

}
