/**
 * JFlex scanner for XSA type and value definitions.
 */

package cks.parser;

import cks.model.Location;

import java.io.Reader;
import java.io.File;
import java.io.IOException;

import java.util.Stack;
import java.math.BigInteger;

%%

%class StreamTokenizer
%implements TokenSource
%public
%final
%apiprivate

%function _scan
%type Token
%throws SyntaxException

%{

	// Builds up a string
	private StringBuilder valueBuilder;

	// Keeps track of that's spread across multiple
	// actions (like for a string literal)
	private StringBuilder textBuilder;
	private char stringEndDelimiter;
	private String fileName;
	private int sectionStartColumn;
	private int sectionStartLine;
	private String indent;

	public StreamTokenizer(Reader in, String fileName)
	{
		this(in);
		this.fileName = fileName;

		this.valueBuilder = new StringBuilder();
		this.textBuilder = new StringBuilder();
		this.line = 1;
		this.column = 1;
		this.indent = "";
	}

	private static String EOL = System.getProperty("line.separator");

	private int line;
	private int column;

	// wrapper to trace the tokens
	public Token get()
		throws IOException, SyntaxException
	{
		Token t = _scan();
		//System.err.println("TOKEN: " + t);
		return t;
	}

	public String getIndent()
	{
		return this.indent;
	}

	private Token use(int type)
	{
		return use(new Token(type));
	}
	
	private Token useSpecial(int type)
	{
		return use("", new Token(type));
	}

	private Token use(Token token)
	{
		return use(yytext(), token);
	}

	private Token use(String string, Token token) {
		token.text = string;
		token.loc = new Location(fileName, line, column);
		skip();
		return token;
	}

	private Token useString()
	{
		Token token = new Token.LitStr(valueBuilder.toString());
		useRemember(token, textBuilder.toString());

		valueBuilder.setLength(0);
		textBuilder.setLength(0);

		skip();
		return token;
	}

	private Token useRemember(Token token, String string)
	{
		token.text = string;
		token.loc = new Location(fileName, sectionStartLine, sectionStartColumn);
		return token;
	}

	private void skip()
	{
		column += yylength();
	}

	private void remember()
	{
		sectionStartLine = line;
		sectionStartColumn = column;
		skip();
	}

	private BigInteger getBigInt(int radix, int start)
	{
		int len = yylength();
		int end = start + len;
		StringBuffer digits = new StringBuffer(len);

		for (int i = start; i < end; i++) {
			char c = yycharat(i);
			if (c != '_') digits.append(c);
		}

		return new BigInteger(digits.toString(), radix);
	}

%}

EOL = \r|\n|\r\n

Any = [^\r\n]

InlineComment = "//" {Any}*

Dec = [_0-9]
Hex = [_A-Fa-f0-9]
Bin = [_01]

DecIntLit = "-"? {Dec}+
BinIntLit = 0b {Bin}+
HexIntLit = 0x {Hex}+

IdentStart = [A-Za-z_]
IdentPart  = [A-Za-z_0-9]
Ident      = {IdentStart}{IdentPart}*

RawDelimiterMarker = @
DelimiterMarker    = %
StandardDelimiter  = [\"']
SingleDelimiter    = [!/|:] | {StandardDelimiter}
PairedDelimiter    = [{\[<]
PairedEndDelimiter = [}\]>]

EndDelimiter   = {SingleDelimiter} | {PairedEndDelimiter}

Delimiter      = {SingleDelimiter} | {PairedDelimiter} | {PairedEndDelimiter} | {DelimiterMarker} | {RawDelimiterMarker}
Escapable      = {Delimiter} | \\

EncloserChars  = "[" | "]" | "{" | "}" | "(" | ")" | "<" | ">"
OtherChars     = "." | "," | ":" | "|" | "=" | "?" | "+" | "/" | "*"
CharTokens     = {EncloserChars} | {OtherChars}

%state STRING BLOCK_COMMENT IMPORT

%%

<YYINITIAL> {

	{Ident}      { return use(new Token(Codes.Ident));    }

	{CharTokens} { return use(yycharat(0)); }
	"</"         { return use(new Token(Codes.CloseTagStart)); }
	"/>"         { return use(new Token(Codes.LeafTagEnd)); }

	{DecIntLit}  { return use(new Token.LitInt(getBigInt(10, 0))); }
	{HexIntLit}  { return use(new Token.LitInt(getBigInt(16, 2))); }
	{BinIntLit}  { return use(new Token.LitInt(getBigInt(2, 2))); }

	^(" "|"\t")* {
		Token tabToken = useSpecial(Codes.Lead);
	}

	(" "|"\t")*{EOL} {
		Token eolToken = useSpecial(Codes.EOL);
		line++;
		indent = "";
		column = 1;
		return eolToken;
	}

	" "+  { skip(); }
	"\t"+ { skip(); }

	{InlineComment} { skip(); }
	"/*"            { remember(); yybegin(BLOCK_COMMENT); }

	{StandardDelimiter} {
		remember();
		stringEndDelimiter = yycharat(0);
		yybegin(STRING);
		textBuilder.append(yytext());
	}

	{DelimiterMarker}{SingleDelimiter} {
		remember();
		stringEndDelimiter = yycharat(1);
		yybegin(STRING);
		textBuilder.append(yytext());
	}

	{DelimiterMarker}{PairedDelimiter} {
		remember();
		stringEndDelimiter = (char) (yycharat(1) + 2);
		yybegin(STRING);
		textBuilder.append(yytext());
	}

	<<EOF>> {
		return useSpecial(Codes.EOF);
	}

	. {
		throw new SyntaxException(use(Codes.Unknown), "unexpected character");
	}
			
}

<STRING> {

	\\n { skip(); valueBuilder.append(EOL); textBuilder.append(yytext()); }
	\\t { skip(); valueBuilder.append('\t');   textBuilder.append(yytext());}
	
	\\{Escapable} {
		skip();
		valueBuilder.append(yycharat(1));
		textBuilder.append(yytext());
	}

	{EOL} {
		throw new SyntaxException(useString(), "unterminated string literal");
	}

	<<EOF>> {
		throw new SyntaxException(useString(), "unterminated string literal");
	}

	{EndDelimiter} {
		textBuilder.append(yytext());
		if (yycharat(0) == stringEndDelimiter) {
			yybegin(YYINITIAL);
			return useString();
		} else {
			valueBuilder.append(yycharat(0));
			skip();
		}
	}

	[^\n\r\t\\] {
		String text = yytext();
		textBuilder.append(yytext());
		valueBuilder.append(yytext());
		skip();
	}

	. {
		textBuilder.append(yytext());
		Token badChar = use(textBuilder.toString(), new Token.LitStr(valueBuilder.toString()));
		valueBuilder.setLength(0);
		textBuilder.setLength(0);
		throw new SyntaxException(badChar, "invalid character in string literal");
	}

}

<BLOCK_COMMENT> {

	[^\*\r\n]*   { skip(); }

	{EOL}       { line++; column = 1; }

	"*"+        { skip(); }

	"*"+"/"     { skip(); yybegin(YYINITIAL); }
	
	<<EOF>>     {
		Token token = new Token(Codes.BlockCommentStart);
		token = useRemember(token, "/#");
		throw new SyntaxException(token, "unterminated block comment");
	}

}

