package tools;

import java.io.File;
import java.io.IOException;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.PrintWriter;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

public class TokenCodeGenerator
{

  private static class Info
  {
    String packageName;
    String className;
    int firstCode = 128;
    List<Code> codes = new ArrayList<Code>(100);
  }

  private static class Code
  {
    String identifier;
    String name;
    boolean displayLiteral;
    boolean displayName;
  }

  public static void main(String[] args)
  {
    if (args.length < 1 || args.length > 2) {
      System.err.println("Usage: COMMAND  <input-file> [output-dir]");
    }

    String inputFileName = args[0];
    File inputFile = new File(inputFileName);


    Info info = new Info();
    String baseName = inputFile.getName();
    info.className = baseName.substring(0, baseName.indexOf('.'));
    try {
      loadInfo(new BufferedReader(new FileReader(inputFile)), info);
    } catch (SyntaxException ex) {
      System.err.println("Error in input: " + ex.getMessage());
      System.exit(2);
    } catch (IOException ex) {
      System.err.println("Error reading from '" + inputFile.getPath() + "': " + ex);
      System.exit(1); return;
    }

    File outputBaseDir = null;
    if (args.length > 1) {
      outputBaseDir = new File(args[1]);
    }
    String packagePath = info.packageName.replace(".", "/");
    File outputDir = new File(outputBaseDir, packagePath);
    if (!outputDir.isDirectory()) outputDir.mkdirs();

    String outputFileName = info.className + ".java";
    File outputFile = new File(outputDir, outputFileName);
    try {
      FileWriter out = new FileWriter(outputFile);
      try {
        generateJava(info, new PrintWriter(out));
      } finally {
        try {
          out.close();
        } catch (IOException ex) {
          System.err.println("Warning: error closing '" + outputFile.getPath() + "': " + ex);
        }
      }
    } catch (IOException ex) {
      System.err.println("Error opening '" + outputFile.getPath() + "' for output: " + ex);
      System.exit(1); return;
    }


  }

  private static class SyntaxException extends Exception
  {
    public SyntaxException(int lineNumber, String message) { super("line " + lineNumber + ": " + message); }
  }

  private static void loadInfo(BufferedReader in, Info info)
    throws IOException, SyntaxException
  {
    int lineNumber = 0;
    boolean displayName = false;
    boolean displayLiteral = true;
    while (true) {
      String original = in.readLine();
      if (original == null) break;
      lineNumber++;

      String line = original.trim();
      if (line.length() == 0) continue;
      if (line.startsWith("#")) continue;

      if (line.startsWith("%class ")) {
        String className = line.substring(7).trim();
        if (className.length() == 0) {
          throw new SyntaxException(lineNumber, "invalid class name '" + className + "'");
        }
        info.className = className;
      } else if (line.startsWith("%package ")) {
        String packageName = line.substring(9).trim();
        if (packageName.length() == 0) {
          throw new SyntaxException(lineNumber, "invalid package name '" + packageName + "'");
        }
        info.packageName = packageName;
      } else if (line.equals("%exact")) {
        displayName = false;
        displayLiteral = true;
      } else if (line.equals("%pattern")) {
        displayName = true;
        displayLiteral = true;
      } else if (line.equals("%named")) {
        displayName = true;
        displayLiteral = false;
      } else if (line.startsWith("%")) {
        throw new SyntaxException(lineNumber, "unknown directive");
      } else {
        Code code = new Code();
        code.displayName = displayName;
        code.displayLiteral = displayLiteral;
        int div = line.indexOf(' ');
        if (div < 0) {
          code.identifier = line;
        } else {
          code.identifier = line.substring(0, div);
          code.name = line.substring(div+1).trim();
        }
        info.codes.add(code);
      }
    }
  }

  private static void generateJava(Info info, PrintWriter out)
  {
    out.println("// automatically generated");
    if (info.packageName != null) {
      out.println("package " + info.packageName + ";");
      out.println();
    }
    out.println("public class " + info.className);
    out.println("{");
    out.println();

    out.println("  public static final int FIRST_CODE = " + info.firstCode + ";");
    out.println("  public static final int LAST_CODE = "  + (info.firstCode + info.codes.size()) + ";");
    out.println();

    out.println("  // token code constants");
    int id = info.firstCode;
    for (Code code : info.codes) {
      out.println("  public static final int " + code.identifier + " = " + id + ";");
      id++;
    }
    out.println();

    out.println("  // token names");
    out.println("  public static final String[] NAMES = {");
    for (Code code : info.codes) {
      String name;
      if (code.name != null) {
        name = escape(code.name);
      } else {
        name = code.identifier;
      }
      out.println("    \"" + name + "\",");
    }
    out.println("  };");
    out.println();

    out.println("  // display flags");

    out.println("  public static final boolean[] DISPLAY_NAME = {");
    for (Code code : info.codes) {
      out.println("    " + code.displayName + ",");
    }
    out.println("  };");

    out.println("  public static final boolean[] DISPLAY_LITERAL = {");
    for (Code code : info.codes) {
      out.println("    " + code.displayLiteral + ",");
    }
    out.println("  };");

    out.println();
    out.println("}");
  }

  private static String escape(String string)
  {
    int pos = string.indexOf('"');
    if (pos < 0) {
      return string;
    } else {
      return string.substring(0, pos) + "\\\"" + escape(string.substring(pos + 1));
    }

  }

}

